searchData={"items":[{"type":"module","title":"CompareChain","doc":"Convenience macros for doing comparisons","ref":"CompareChain.html"},{"type":"module","title":"Valid expressions - CompareChain","doc":"Valid expressions for `compare?/1` and `compare?/2` follow these three rules:","ref":"CompareChain.html#module-valid-expressions"},{"type":"module","title":"1. A comparison operator like `<` must be present. - CompareChain","doc":"At least one of these must be included: `<`, `>`, `<=`, `>=`, `==`, `!=`,\n`===`, or `!===`. So this is valid:\n\n    compare?(1 < 2 < 3)\n\nbut this is not:\n\n    compare?(true)\n\nbecause it does not contain any comparisons.","ref":"CompareChain.html#module-1-a-comparison-operator-like-must-be-present"},{"type":"module","title":"2. All arguments to boolean operators must also be valid expressions. - CompareChain","doc":"The boolean operators `and`, `or`, and `not` are allowed in expressions so\nlong as all of their arguments (eventually) contain a comparison. So this is\nvalid:\n\n    compare?(1 < 2 < 3 and 4 < 5)\n\nas is this:\n\n    compare?(not (not 1 < 2 < 3))\n\nbut this is not:\n\n    compare?(1 < 2 < 3 and true)\n\nbecause the right side of `and` fails to contain a comparison. This\nexpression can be refactored to be valid by moving the non-comparison branch\noutside `compare?/1` like so:\n\n    compare?(1 < 2 < 3) and true","ref":"CompareChain.html#module-2-all-arguments-to-boolean-operators-must-also-be-valid-expressions"},{"type":"module","title":"3. The root operator of an expression must be a comparison or a boolean. - CompareChain","doc":"So this is not valid:\n\n    compare?(my_function(a < b), Date)\n\nbecause its root operator is `my_function/1`. This expression can be\nrefactored to be valid by moving `compare?/2` inside `my_function/1` like so:\n\n    my_function(compare?(a < b, Date))\n\nWe restrict expressions in this fashion so we can guarantee that `compare?/1`\nand `compare?/2` will always return a boolean.\n\nAlso note that arguments to _comparisons_ may be arbitrarily complicated:\n\n    compare?(a < Date.utc_today(), Date)","ref":"CompareChain.html#module-3-the-root-operator-of-an-expression-must-be-a-comparison-or-a-boolean"},{"type":"macro","title":"CompareChain.compare?/1","doc":"Macro that performs chained comparison with operators like `<`.\n\nYou may also include the boolean operators `and`, `or`, and `not` in the\nexpression so long as all their arguments all (eventually) contain\ncomparisons. See the moduledoc for more details.\n\nFor a version that also does semantic comparison, see: `compare?/2`.","ref":"CompareChain.html#compare?/1"},{"type":"macro","title":"Examples - CompareChain.compare?/1","doc":"Chained comparison:\n\n    iex> import CompareChain\n    iex> compare?(1 < 2 < 3)\n    true\n\nComparisons joined by logical operators:\n\n    iex> import CompareChain\n    iex> compare?(1 >= 2 >= 3 or 4 >= 5 >= 6)\n    false","ref":"CompareChain.html#compare?/1-examples"},{"type":"macro","title":"Warnings and errors - CompareChain.compare?/1","doc":">","ref":"CompareChain.html#compare?/1-warnings-and-errors"},{"type":"macro","title":"Comparing structs will warn {: .warning} - CompareChain.compare?/1","doc":">\n> Expressions which compare matching structs like:\n>\n>     iex> compare?(~D[2017-03-31] < ~D[2017-04-01])\n>     false\n>\n> Will result in a warning:\n>\n> ```plain\n> ... [warning] Performing structural comparison on matching structs.\n>\n> Did you mean to use `compare?/2`?\n>\n>   compare?(~D[2017-03-31] < ~D[2017-04-01], Date)\n> ```\n>\n> You probably want to use `compare?/2`, which does semantic comparison,\n> instead.\n\n>","ref":"CompareChain.html#compare?/1-comparing-structs-will-warn-warning"},{"type":"macro","title":"Invalid expressions will raise {: .error} - CompareChain.compare?/1","doc":">\n> See the section on valid expressions in the moduledoc for details.","ref":"CompareChain.html#compare?/1-invalid-expressions-will-raise-error"},{"type":"macro","title":"CompareChain.compare?/2","doc":"Macro that performs chained, semantic comparison with operators like `<` by\nrewriting the expression using the `compare/2` function defined by the\nprovided module.\n\nThis is like how you can provide a module as the second argument to\n`Enum.sort/2` when you need to sort items semantically.\n\nYou may also include the boolean operators `and`, `or`, and `not` in the\nexpression so long as all their arguments all (eventually) contain\ncomparisons. See the moduledoc for more details.\n\nFor a version that does chained comparison using the normal `<` operators,\nsee: `compare?/1`.","ref":"CompareChain.html#compare?/2"},{"type":"macro","title":"Examples - CompareChain.compare?/2","doc":"Semantic comparison:\n\n    iex> import CompareChain\n    iex> a = ~D[2017-03-31]\n    iex> b = ~D[2017-04-01]\n    iex> compare?(a   #### Semantic vs. Structural Comparison Differences {: .info}\n>\n> In the above example, `compare?(a   own, `a   important to not use comparison operators on structs directly.** The answer\n> is not what you would expect.\n>\n> _Trivia!_ If you're curious, `b` comes before `a` because in term ordering,\n> maps of equal size are compared key by key in ascending order. In this case,\n> `:day` is the first key (due to ASCII byte ordering) where `a` and `b`\n> differ. Since `a.day == 31` and `b.day == 1`, we have `b   import CompareChain\n    iex> a = ~D[2017-03-31]\n    iex> b = ~D[2017-04-01]\n    iex> c = ~D[2017-04-02]\n    iex> compare?(a   import CompareChain\n    iex> a = ~T[15:00:00]\n    iex> b = ~T[16:00:00]\n    iex> c = ~T[17:00:00]\n    iex> compare?(a   c, Time)\n    false\n\nMore complex expressions:\n\n    iex> import CompareChain\n    iex> compare?(%{a: ~T[16:00:00]}.a <= ~T[17:00:00], Time)\n    true\n\nCustom module:\n\n    iex> import CompareChain\n    iex> defmodule AlwaysGreaterThan do\n    iex>   def compare(_left, _right), do: :gt\n    iex> end\n    iex> compare?(1 > 2 > 3, AlwaysGreaterThan)\n    true","ref":"CompareChain.html#compare?/2-examples"},{"type":"macro","title":"Warnings and errors - CompareChain.compare?/2","doc":">","ref":"CompareChain.html#compare?/2-warnings-and-errors"},{"type":"macro","title":"Using the \"strict\" operators will warn {: .warning} - CompareChain.compare?/2","doc":">\n> Expressions which include either `===` or `!==` like:\n>\n>     iex> compare?(~D[2017-03-31] !== ~D[2017-04-01], Date)\n>     true\n>\n> Will result in a warning:\n>\n> ```plain\n> ... [warning] Performing semantic comparison using either: `===` or `!===`.\n> This is reinterpreted as `==` or `!=`, respectively.\n> ```\n>\n> These operators have no additional meaning over `==` and `!=` when doing\n> semantic comparison.\n\n>","ref":"CompareChain.html#compare?/2-using-the-strict-operators-will-warn-warning"},{"type":"macro","title":"Invalid expressions will raise {: .error} - CompareChain.compare?/2","doc":">\n> See the section on valid expressions in the moduledoc for details.","ref":"CompareChain.html#compare?/2-invalid-expressions-will-raise-error"},{"type":"extras","title":"CompareChain","doc":"# CompareChain\n\n**Chained semantic comparisons for Elixir.**\n\n[![Package](https://img.shields.io/hexpm/v/compare_chain?logo=elixir&style=for-the-badge)](https://hex.pm/packages/compare_chain)\n[![Downloads](https://img.shields.io/hexpm/dt/compare_chain?logo=elixir&style=for-the-badge)](https://hex.pm/packages/compare_chain)\n[![Build](https://img.shields.io/github/actions/workflow/status/CargoSense/compare_chain/ci.yml?branch=main&logo=github&style=for-the-badge)](https://github.com/CargoSense/compare_chain/actions/workflows/ci.yml)","ref":"readme.html"},{"type":"extras","title":"Key Features - CompareChain","doc":"CompareChain provides convenience macros for:\n\n- chained comparisons (`a  `, `<=`, `>=`, `==`, `!=`, `===`, and `!==`)\n- combinations (`and`, `or`, and `not`)","ref":"readme.html#key-features"},{"type":"extras","title":"Installation - CompareChain","doc":"Add `compare_chain` to your project's dependencies in `mix.exs` and run `mix deps.get`:\n\n```elixir\ndef deps do\n  [\n    {:compare_chain, \"~> 0.5\"}\n  ]\nend\n```","ref":"readme.html#installation"},{"type":"extras","title":"Usage - CompareChain","doc":"Import CompareChain to enable access to `CompareChain.compare?/1` and `CompareChain.compare?/2`:\n\n```elixir\niex> import CompareChain\n\n# Chained comparisons\niex> compare?(1 < 2 < 3)\ntrue\n\n# Semantic comparisons\niex> compare?(~D[2017-03-31] < ~D[2017-04-01], Date)\ntrue\n\n# Chained semantic comparisons\niex> compare?(~D[2017-03-31] < ~D[2017-04-01] < ~D[2017-04-02], Date)\ntrue\n\n# Semantic comparisons with logical operators\niex> compare?(~T[16:00:00] <= ~T[16:00:00] and not (~T[17:00:00] <= ~T[17:00:00]), Time)\nfalse\n\n# Complex expressions\niex> compare?(%{a: ~T[16:00:00]}.a <= ~T[17:00:00], Time)\ntrue\n```\n\nSee [CompareChain on HexDocs](https://hexdocs.pm/compare_chain) for more.","ref":"readme.html#usage"},{"type":"extras","title":"Background - CompareChain","doc":"Say you have an interval of time bounded by two `%Date{}` structs, `start_date` and `end_date`, and you want to know whether or not a third `date` falls within that interval. How would you write this in Elixir?\n\n```elixir\nDate.compare(start_date, date) == :lt and\n  Date.compare(date, end_date) == :lt\n```\n\nThe above code is verbose, somewhat hard to read, and potentially incorrect (though not obviously so). What if `date` is considered \"within\" the interval inclusive of the `start_date` or `end_date`? To include the bounds in the comparison, you'd instead write the expression like this:\n\n```elixir\nDate.compare(start_date, date) != :gt and\n  Date.compare(date, end_date) != :gt\n\n# …or, even more verbosely:\nDate.compare(start_date, date) in [:lt, :eq] and\n  Date.compare(date, end_date) in [:lt, :eq]\n```\n\nTo spot the difference between these two cases, you must keep in mind:\n\n- the order of the arguments passed to `Date.compare/2`,\n- the specific comparison operators for each clause (`==` vs. `!=`), and\n- the specific comparison atoms for each clause (`:lt` vs. `:gt`).\n\nContrast this example with equivalent Python code:\n\n```python\n# excluding bounds\nstart_date < date < end_date\n\n# including bounds\nstart_date <= date <= end_date\n```\n\nMuch easier to read! Why can't you write this in Elixir? Two reasons:\n\n1. Structural comparison operators\n2. Chained vs. nested comparisons","ref":"readme.html#background"},{"type":"extras","title":"Structural Comparison Operators - CompareChain","doc":"Operators like `<` do _structural_ comparison (instead of _semantic_ comparison). From the [`Kernel` docs](https://hexdocs.pm/elixir/Kernel.html#module-structural-comparison):\n\n> …**comparisons in Elixir are structural**, as it has the goal of comparing data types as efficiently as possible to create flexible and performant data structures. This distinction is specially important for functions that provide ordering, such as `>/2`, `</2`, `>=/2`, `<=/2`, `min/2`, and `max/2`. For example:\n>\n> ```elixir\n> ~D[2017-03-31] > ~D[2017-04-01]\n> ```\n>\n> will return `true` because structural comparison compares the `:day` field before `:month` or `:year`. In order to perform semantic comparisons, the relevant data-types provide a `compare/2` function, such as `Date.compare/2`:\n>\n> ```elixir\n> iex> Date.compare(~D[2017-03-31], ~D[2017-04-01])\n> :lt\n> ```\n\nIn other words, although `~D[2017-03-31] > ~D[2017-04-01]` is valid code, it does _not_ tell you if `~D[2017-03-31]` is a later date than `~D[2017-04-01]` as you might expect.\nInstead, you'd use `Date.compare/2`.","ref":"readme.html#structural-comparison-operators"},{"type":"extras","title":"Chained vs. Nested Comparisons - CompareChain","doc":"Additionally, even if `~D[2017-03-31] > ~D[2017-04-01]` did semantic comparison, you still couldn't write the interval check like you do in Python. In Python, an expression like `1 < 2 < 3` is syntactic sugar for `(1 < 2) and (2 < 3)` (a series of \"chained\" expressions).\n\nElixir doesn't provide an equivalent syntactic sugar. Instead, `1 < 2 < 3` is evaluated as `(1 < 2) < 3` (a series of \"nested\" expressions). `(1 < 2) < 3` evaluates to `true < 3` which is _probably_ not what you want!","ref":"readme.html#chained-vs-nested-comparisons"},{"type":"extras","title":"A Solution! - CompareChain","doc":"CompareChain addresses these complexities with the macro `CompareChain.compare?/2`:\n\n```elixir\nimport CompareChain\n\n# excluding bounds\ncompare?(start_date < date < end_date, Date)\n\n# including bounds\ncompare?(start_date <= date <= end_date, Date)\n```\n\n`CompareChain.compare?/2` rewrites these expressions as:\n\n```elixir\n# excluding bounds\nDate.compare(start_date, date) == :lt and\n  Date.compare(date, end_date) == :lt\n\n# including bounds\nDate.compare(start_date, date) != :gt and\n  Date.compare(date, end_date) != :gt\n```\n\nYour code is more readable while remaining correct!\n\n`CompareChain.compare?/1` also enables chained comparison using the structural operators:\n\n```elixir\ncompare?(1 < 2 < 3)\n```","ref":"readme.html#a-solution"},{"type":"extras","title":"Acknowledgements - CompareChain","doc":"Thanks to [Ben Wilson](https://github.com/benwilson512) and [Michael Crumm](https://github.com/mcrumm) for the helpful discussions and their guidance!\n\nThanks as well to the folks who participated in the [elixir-lang-core](https://groups.google.com/g/elixir-lang-core) discussion, particularly Cliff whose [idea](https://groups.google.com/g/elixir-lang-core/c/W2TeQm5r1H4/m/ctVuN_woBgAJ) I shamelessly built off.","ref":"readme.html#acknowledgements"},{"type":"extras","title":"License - CompareChain","doc":"CompareChain is freely available under the [MIT License](https://opensource.org/licenses/MIT).","ref":"readme.html#license"},{"type":"extras","title":"CHANGELOG","doc":"# CHANGELOG","ref":"changelog.html"},{"type":"extras","title":"v0.6.0 (2025-01-09) - CHANGELOG","doc":"* Drop official support for Elixir 1.13 and 1.14 (though they should still work)\n  * Currently support Elixir 1.15 to 1.18\n  * Change docs so they no longer warn about a doctest\n  * Miscellanea: renamed LISENCE file, added CODEOWNERS file, etc.","ref":"changelog.html#v0-6-0-2025-01-09"},{"type":"extras","title":"v0.5.0 (2024-05-22) - CHANGELOG","doc":"* Allow `===` and `!===` in expressions\n  * Improve documentation\n  * Fix bug with certain operation chains:\n    ```\n    compare?(1 < 2 != 3 < 4) #=> true\n    compare?(1 < 2 != 3 > 4) #=> true (wrong!)\n    ```\n  * Improve error message and documentation for invalid expressions\n  * [BREAKING] all branches of `and`, `or` and `not` must now contain comparisons\n    * Example: `compare?(1 < 2 and true)` used to be ok but is no longer\n      allowed because the right argument to `and` doesn't contain a comparison.","ref":"changelog.html#v0-5-0-2024-05-22"},{"type":"extras","title":"v0.4.0 (2023-09-10) - CHANGELOG","doc":"* Warn when `compare?/1` is used on a struct","ref":"changelog.html#v0-4-0-2023-09-10"},{"type":"extras","title":"v0.3.0 (2023-01-28) - CHANGELOG","doc":"* Allow `==` and `!=` in expressions\n  * Allow `Elixir >= 1.13.0`\n    * `Macro.prewalker` was introduced in `1.13.0`","ref":"changelog.html#v0-3-0-2023-01-28"},{"type":"extras","title":"v0.2.0 (2022-11-05) - CHANGELOG","doc":"* Allow `not` in expressions","ref":"changelog.html#v0-2-0-2022-11-05"},{"type":"extras","title":"v0.1.0 (2022-11-03) - CHANGELOG","doc":"* Initial release","ref":"changelog.html#v0-1-0-2022-11-03"}],"proglang":"elixir","content_type":"text/markdown","producer":{"name":"ex_doc","version":"0.36.1"}}